@@ -12,7 +12,7 @@ import re
 import base64
 import hashlib
 from io import StringIO, IOBase
-from typing import Dict, List
+from typing import Dict, List, Tuple
 
 
 logger = logging.getLogger(__name__)
@@ -138,7 +138,7 @@ def get_entity_value_by_state_type_kind_value_type_kind(entity, state_type,
         value_kind: String. The value type of the value.
 
     Returns:
-        A dict object representing the value if it exits. Ot.herwise it
+        A dict object representing the value if it exits. Otherwise it
         returns None
 
     """
@@ -155,6 +155,31 @@ def get_entity_value_by_state_type_kind_value_type_kind(entity, state_type,
                         break
     return value
 
+
+def get_entity_label_by_label_type_kind(entity: dict, label_type: str, label_kind: str) -> str:
+    """Get a label from an acas entity dict object.
+
+    Gets a specific label from an acas entity dict object.
+
+    Args:
+        entity: Any ACAS entity (protocol, experiment, analysis_group,
+        treatment_group, container...etc.)
+        label_type: String. The label type of the value.
+        label_kind: String. The label kind of the value.
+
+    Returns:
+        A dict object representing the label if it exits. Otherwise it
+        returns None
+
+    """
+    label = None
+    for l in entity["lsLabels"]:
+        if (not l["deleted"] and not l["ignored"] and
+                l["lsType"] == label_type and l["lsKind"] == label_kind):
+            label = l
+            break
+    return label
+
 def sdf_iterator(iteratable):
     data = []
     for line in iteratable:
@@ -336,6 +361,141 @@ class client():
         resp.raise_for_status()
         return resp.json()
 
+    def add_files_to_lot(self, lot_corp_name, files):
+        """Attach files to a lot.
+
+        Attach files to a lot by passing a list of files and the lot's corp name.
+
+        Args:
+            lot_corp_name: The corp name of the lot to attach the files to.
+            files: An array of dicts with the following keys
+                "file" (Path): The file path to upload
+                "type" (str): The type of file to upload
+                "writeup" (str): The writeup for the file
+        
+        Returns: A dict with errors array and metalot dict object with the files attached to the lot
+
+        """
+        
+        # Get the meta lot to make sure it exists
+        meta_lot = self.get_meta_lot(lot_corp_name)
+
+        # Upload the files to cmpd reg
+        uploaded_files = self._upload_cmpdreg_files(lot_corp_name, files)
+
+        # Attach the files to the lot
+        for uploaded_file in uploaded_files:
+            meta_lot['fileList'].append(uploaded_file)
+
+        # Update the meta lot
+        meta_lot_save_response = self.save_meta_lot(meta_lot)
+
+        return meta_lot_save_response
+
+    def add_file_to_lot(self, lot_corp_name, file, file_type, writeup=None):
+        """Add a file to a lot
+
+        Uploads and attaches a file to a lot
+
+        Args:
+            lot_corp_name (str): The corp name of the lot to attach the files to.
+            file (Path): The file path to upload
+            type (str): The type of file to upload
+            writeup (str): The writeup for the file
+
+        Returns: A dict with errors array and metalot dict object with the files attached to the lot
+
+        """
+        return self.add_files_to_lot(lot_corp_name, [{"file": file, "file_type": file_type, "writeup": writeup}])
+
+    def _upload_cmpdreg_files(self, lot_corp_name, files):
+        """Upload a list of files to CmpdReg.
+
+        Pass an array of files to ACAS and upload them to the creg server
+
+        Args:
+            lot_corp_name (str): The lot corp name to upload the file to
+            files: An array of dicts with the following keys
+                "file" (Path): The file path to upload
+                "type" (str): The type of file to upload
+                "writeup" (str): The writeup for the file
+
+        Returns:
+            An object of responses from ACAS in the form:
+
+        .. code-block:: python
+
+            [
+                {
+                    "description": "HPLC",
+                    "ie": true,
+                    "name": "dummy.pdf",
+                    "size": 13264,
+                    "subdir": "CMPD-0000002-001",
+                    "type": "application/pdf",
+                    "uploaded": true,
+                    "url": "getFile?file
+                }
+            ]
+        """
+        filesToUpload = ()
+        for file_dict in files:
+            filesToUpload = filesToUpload + (('file[]', (file_dict["file"].name, file_dict["file"].open('rb'), 'application/octet-stream')),)
+            filesToUpload = filesToUpload + (('description[]', (None, file_dict["file_type"])),)
+            # If there is no writeup, or if we set the writeup to empty quotes "" the multipart form service in acas will fail
+            # Setting a single space character sets the writeup to "" in the database
+            if "writeup" not in file_dict or file_dict["writeup"] is None:
+                writeup = " "
+            else:
+                writeup = file_dict["writeup"]
+            filesToUpload = filesToUpload + (('writeup[]', (None, writeup)),)
+
+        filesToUpload = filesToUpload + (('subdir', (None, lot_corp_name)),)
+        filesToUpload = filesToUpload + (('ie', (None, 'true')),)
+
+        headers = {
+            'Accept': 'application/json'
+        }
+        resp = self.session.post("{}/cmpdreg/filesave".format(self.url),
+                                 headers=headers,
+                                 files=filesToUpload)
+
+        resp.raise_for_status()
+        return resp.json()
+
+    def _upload_cmpd_reg_file(self, lot_corp_name, file, file_type, writeup=None):
+        """Upload a file to CmpdReg.
+
+        Upload a file to a specific lot corp name folder in cmpdreg
+        Note: This does not attach the file to the specific lot it just uploads the file to that folder
+
+        Args:
+            lot_corp_name (str): The lot corp name to upload the file to
+            file (Path): The file path to upload
+            type (str): The type of file to upload
+            writeup (str): The writeup for the file
+
+        Returns:
+            An object of responses from ACAS in the form:
+        .. code-block:: python
+
+            [
+                {
+                    "description": "HPLC",
+                    "ie": true,
+                    "name": "dummy.pdf",
+                    "size": 13264,
+                    "subdir": "CMPD-0000002-001",
+                    "type": "application/pdf",
+                    "uploaded": true,
+                    "url": "getFile?fileUrl=%2Fhome%2Frunner%2Fbuild%2FprivateUploads%2Fcmpdreg%2FnoteBookFiles%2FnoteBook%2FCMPD-0000002-001%2Fdummy.pdf",
+                    "writeup": "Nothing"
+                }
+            ]
+        """
+        response = self._upload_cmpdreg_files(lot_corp_name, [{"file": file, "file_type": file_type, "writeup": writeup}])
+        return response[0]
+
     def get_meta_lot(self, lot_corp_name):
         """Get metalot by lot corp name
          Granted read permission on a lot if one of these is true:
@@ -2134,4 +2294,32 @@ en array of protocols
                 aliases.append(new_alias_obj)
         
         # Persist the update
-        self.save_meta_lot(meta_lot)
\ No newline at end of file
+        self.save_meta_lot(meta_lot)
+    
+    def edit_parent(self, parent, dry_run=True) -> Tuple[bool, Dict]:
+        """Makes changes to an existing parent.
+        Returns (status, data) where `status` is a bool representing success (True) or failure (False)
+        `data` is a dict of supporting information. In case of failure it contains a list of duplicates
+        In case of success it contains a list of affected lots
+        """
+        # Validate
+        resp = self.session.post("{}/cmpdreg/validateParent".format(self.url),
+                                 headers={'Content-Type': "application/json"},
+                                 data=json.dumps(parent))
+        resp.raise_for_status()
+        resp_data = resp.json()
+        success = True
+        # Check whether the response just has a list of affected lots or whether
+        # It has a list of dupeParents
+        if 'parentUnique' in resp_data and not resp_data['parentUnique']:
+            success = False
+        # If just doing validation, or if validation failed, return response
+        if dry_run or not success:
+            return success, resp.json()
+        # Otherwise continue to updateParent
+        resp = self.session.post("{}/cmpdreg/updateParent".format(self.url),
+                                 headers={'Content-Type': "application/json"},
+                                 data=json.dumps(parent))
+        resp.raise_for_status()
+        resp_data = resp.json()
+        return success, resp_data
\ No newline at end of file
